/*
  +----------------------------------------------------------------------+
  | Hidef                                                                |
  +----------------------------------------------------------------------+
  | Copyright (c) 2008 The PHP Group                                     |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Gopal Vijayaraghavan <gopalv@php.net>                       |
  +----------------------------------------------------------------------+
*/

/* $Id: frozenarray.c 299156 2010-05-08 20:57:02Z kalle $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "zend_globals.h"
#include "zend_interfaces.h"
#include "php_streams.h"
#include "ext/standard/php_var.h"
#include "ext/standard/php_smart_str.h"

#include "php_hidef.h"
#include "frozenarray.h"

/* {{{ refcount macros */
#ifndef Z_ADDREF_P
#define Z_ADDREF_P(pz)                (pz)->refcount++
#define Z_ADDREF_PP(ppz)              Z_ADDREF_P(*(ppz))
#endif

#ifndef Z_DELREF_P
#define Z_DELREF_P(pz)                (pz)->refcount--
#define Z_DELREF_PP(ppz)              Z_DELREF_P(*(ppz))
#endif
/* }}} */

/* {{{ arginfo static macro */
#if PHP_MAJOR_VERSION > 5 || PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION >= 3
#define HIDEF_ARGINFO_STATIC
#else
#define HIDEF_ARGINFO_STATIC static
#endif
/* }}} */

/* {{{ zend_parse_parameters_none() macro */
#ifndef zend_parse_parameters_none
#define zend_parse_parameters_none() zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "")
#endif
/* }}} */


static zend_class_entry* frozen_array_ce = NULL;

static zend_object_handlers frozen_array_object_handlers;

PHP_HIDEF_API zend_class_entry* frozen_array_get_ce()
{
	return frozen_array_ce;
}

#define FROZEN_ME(func, arg_info, flags) \
	PHP_ME(FrozenArray, func, arg_info, flags)

#define FROZEN_METHOD(func) \
	PHP_METHOD(FrozenArray, func)

static frozen_array_object* frozen_array_object_new(zend_class_entry *ce TSRMLS_DC);
static zend_object_value frozen_array_register_object(frozen_array_object* intern TSRMLS_DC);

static void frozen_array_object_clone(void *object, void **clone_ptr TSRMLS_DC);

static void frozen_array_object_dtor(void *object, zend_object_handle handle TSRMLS_DC);
static void frozen_array_object_free_storage(void *object TSRMLS_DC);

static frozen_array_object* frozen_array_object_new(zend_class_entry *ce TSRMLS_DC) 
{
	frozen_array_object* intern = NULL;

	intern = ecalloc(1, sizeof(frozen_array_object));

#if (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION == 1 && PHP_RELEASE_VERSION > 2) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION > 1) || (PHP_MAJOR_VERSION > 5)
	zend_object_std_init(&intern->zo, ce TSRMLS_CC);
#else
	ALLOC_HASHTABLE(intern->zo.properties);
	zend_hash_init(intern->zo.properties, 0, NULL, ZVAL_PTR_DTOR, 0);

	intern->zo.ce = ce;
	intern->zo.guards = NULL;
#endif

	return intern;
}

static zend_object_value frozen_array_register_object(frozen_array_object* intern TSRMLS_DC) 
{

	zend_object_value rv;

	rv.handle = zend_objects_store_put(intern, 
					frozen_array_object_dtor, 
					(zend_objects_free_object_storage_t)frozen_array_object_free_storage, 
					frozen_array_object_clone TSRMLS_CC);
	
	/* does not support (EG(ze1_compatibility_mode)) */

	rv.handlers = (zend_object_handlers *) &frozen_array_object_handlers;

	return rv;
}

static void frozen_array_object_clone(void *object, void **clone_ptr TSRMLS_DC)
{
	frozen_array_object *intern = (frozen_array_object*) object;
	frozen_array_object *clone = NULL;

	clone = frozen_array_object_new(intern->zo.ce TSRMLS_CC);

	/* move data across */

	*clone_ptr = clone;
}


static void frozen_array_object_dtor(void *object, zend_object_handle handle TSRMLS_DC)
{
	/* nothing to do yet */
}

static void frozen_array_object_free_storage(void *object TSRMLS_DC)
{
	frozen_array_object *intern = (frozen_array_object*) object;

#if (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION == 1 && PHP_RELEASE_VERSION > 2) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION > 1) || (PHP_MAJOR_VERSION > 5)
	zend_object_std_dtor(&intern->zo TSRMLS_CC);
#else
	if (intern->zo.guards)
	{
		zend_hash_destroy(intern->zo.guards);
		FREE_HASHTABLE(intern->zo.guards);
	}

	if (intern->zo.properties)
	{
		zend_hash_destroy(intern->zo.properties);
		FREE_HASHTABLE(intern->zo.properties);
	}
#endif

	/* other cleanups (here or in dtor, check docs) */
	
	if(intern->data) 
	{
		efree(intern->data);
	}
	efree(intern);
}

PHP_HIDEF_API zend_object_value frozen_array_new(zend_class_entry *ce TSRMLS_DC)
{
	frozen_array_object *intern = frozen_array_object_new(ce TSRMLS_CC);
	return frozen_array_register_object(intern TSRMLS_CC);
}

zval* frozen_array_wrap_zval(zval* src TSRMLS_DC)
{
	zval* dst;
	frozen_array_object *intern;

	MAKE_STD_ZVAL(dst);

	if(Z_TYPE_P(src) != IS_ARRAY)
	{
		*dst = *src;
		if(Z_TYPE_P(src) == IS_STRING)
		{
			Z_STRVAL_P(dst) = emalloc(sizeof(char)*Z_STRLEN_P(src)+1);
			memcpy(Z_STRVAL_P(dst), Z_STRVAL_P(src), Z_STRLEN_P(src)+1);
		}
		return dst;
	}

	intern = frozen_array_object_new(frozen_array_ce TSRMLS_CC);

	intern->data = emalloc(sizeof(zval));
	*intern->data = *src;

	dst->type = IS_OBJECT;
	dst->value.obj = frozen_array_register_object(intern TSRMLS_CC);

	return dst;
}

FROZEN_METHOD(offsetExists)
{
	zval *zindex;

	zval *object = getThis();
	frozen_array_object *intern = (frozen_array_object*)zend_object_store_get_object(object TSRMLS_CC);

	/* check data */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &zindex) == FAILURE)
	{
		return;
	}

	if(Z_TYPE_P(zindex) != IS_STRING && Z_TYPE_P(zindex) != IS_LONG)
	{
		convert_to_string(zindex);
	}

	if(intern->data && (Z_TYPE_P(intern->data) == IS_ARRAY))
	{
		if(Z_TYPE_P(zindex) == IS_STRING)
		{
			if(zend_hash_exists(Z_ARRVAL_P(intern->data), Z_STRVAL_P(zindex), Z_STRLEN_P(zindex)+1))
			{
				RETURN_BOOL(1);
			}
		}
		else if(Z_TYPE_P(zindex) == IS_LONG)
		{
			if(zend_hash_index_exists(Z_ARRVAL_P(intern->data), Z_LVAL_P(zindex)))
			{
				RETURN_BOOL(1);
			}
		}
	}

	RETURN_BOOL(0);
}

FROZEN_METHOD(offsetGet)
{
	zval *zindex;
	zval *retval;
	zval **hentry;
	zval *object = getThis();
	frozen_array_object *intern = (frozen_array_object*)zend_object_store_get_object(object TSRMLS_CC);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &zindex) == FAILURE)
	{
		return;
	}

	if(Z_TYPE_P(zindex) != IS_STRING && Z_TYPE_P(zindex) != IS_LONG)
	{
		convert_to_string(zindex);
	}

	if(intern->data && (Z_TYPE_P(intern->data) == IS_ARRAY))
	{
		if(Z_TYPE_P(zindex) == IS_STRING)
		{
			if(zend_hash_find(Z_ARRVAL_P(intern->data), Z_STRVAL_P(zindex), Z_STRLEN_P(zindex)+1, (void**)(&hentry)) != FAILURE)
			{
				retval = frozen_array_wrap_zval(hentry[0] TSRMLS_CC);
				RETURN_ZVAL(retval, 0, 1);
			}
		}
		else if(Z_TYPE_P(zindex) == IS_LONG)
		{
			if(zend_hash_index_find(Z_ARRVAL_P(intern->data), Z_LVAL_P(zindex), (void**)(&hentry)) != FAILURE)
			{
				retval = frozen_array_wrap_zval(hentry[0] TSRMLS_CC);
				RETURN_ZVAL(retval, 0, 1);
			}
		}
	}

	RETURN_NULL();
}

FROZEN_METHOD(offsetSet)
{
	if (zend_parse_parameters_none() == FAILURE)
	{
		return;
	}

	/* throw exception */
}
FROZEN_METHOD(offsetUnset)
{
	if (zend_parse_parameters_none() == FAILURE)
	{
		return;
	}

	/* throw exception */
}

FROZEN_METHOD(__toString)
{
	smart_str str = {0,};
	zval *object = getThis();
	frozen_array_object *intern = (frozen_array_object*)zend_object_store_get_object(object TSRMLS_CC);
	HashTable *data = Z_ARRVAL_P(intern->data);

	if (zend_parse_parameters_none() == FAILURE)
	{
		return;
	}

	smart_str_appends(&str, "FrozenArray[");
	smart_str_append_unsigned(&str, data->nNumOfElements);
	smart_str_appends(&str, "]");
	smart_str_0(&str);

	RETVAL_STRING(str.c, 1);

	smart_str_free(&str);
	return;
}

FROZEN_METHOD(__construct)
{
	if (zend_parse_parameters_none() == FAILURE)
	{
		return;
	}

	/* empty */
}

HIDEF_ARGINFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_void, 0, 0, 0)
ZEND_END_ARG_INFO()

HIDEF_ARGINFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_offsetGet, 0, 0, 1)
	ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO()

HIDEF_ARGINFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_offsetSet, 0, 0, 2)
	ZEND_ARG_INFO(0, index)
	ZEND_ARG_INFO(0, newval)
ZEND_END_ARG_INFO()

static zend_function_entry frozen_array_functions[] = {
	PHP_ME(FrozenArray, __construct,  arginfo_void,      ZEND_ACC_PRIVATE|ZEND_ACC_FINAL)
	PHP_ME(FrozenArray, offsetExists, arginfo_offsetGet, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	PHP_ME(FrozenArray, offsetGet,    arginfo_offsetGet, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	PHP_ME(FrozenArray, offsetSet,    arginfo_offsetSet, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	PHP_ME(FrozenArray, offsetUnset,  arginfo_offsetGet, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	PHP_ME(FrozenArray, __toString,   arginfo_void,      ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL}
};

/* {{{ frozen array iterator */
typedef struct _frozen_array_it 
{
	zend_user_iterator	intern;
	frozen_array_object	*object;
	HashPosition pos;
	zval *current;
} frozen_array_it;

static void frozen_array_it_dtor(zend_object_iterator *iter TSRMLS_DC)
{
	frozen_array_it *iterator = (frozen_array_it*)(iter);
	Z_DELREF_PP((zval**)(&iterator->intern.it.data));
	if(iterator->current) zval_ptr_dtor(&iterator->current);
	efree(iterator);
}


static int frozen_array_it_valid(zend_object_iterator *iter TSRMLS_DC)
{
	frozen_array_it *iterator = (frozen_array_it*)(iter);
	frozen_array_object *obj = iterator->object;
	return zend_hash_has_more_elements_ex(Z_ARRVAL_P(obj->data), &(iterator->pos));
}

static void frozen_array_it_get_current_data(zend_object_iterator *iter, zval ***data TSRMLS_DC)
{
	zval **hentry;

	frozen_array_it *iterator = (frozen_array_it*)(iter);
	frozen_array_object *obj = iterator->object;

	if(zend_hash_get_current_data_ex(Z_ARRVAL_P(obj->data), (void**)(&hentry), &(iterator->pos)) != FAILURE)
	{
		if(iterator->current) zval_ptr_dtor(&iterator->current);
		iterator->current = frozen_array_wrap_zval(hentry[0] TSRMLS_CC);
		*data = &(iterator->current);
	}
}

static int frozen_array_it_get_current_key(zend_object_iterator *iter, char **str_key, uint *str_key_len, ulong *int_key TSRMLS_DC)
{
	frozen_array_it *iterator = (frozen_array_it*)(iter);
	frozen_array_object *obj = iterator->object;

	return zend_hash_get_current_key_ex(Z_ARRVAL_P(obj->data), str_key, str_key_len, int_key, 1, &(iterator->pos));
}

static void frozen_array_it_move_forward(zend_object_iterator *iter TSRMLS_DC)
{
	frozen_array_it *iterator = (frozen_array_it*)(iter);
	frozen_array_object *obj = iterator->object;
	zend_hash_move_forward_ex(Z_ARRVAL_P(obj->data), &(iterator->pos));
}

static void frozen_array_it_rewind(zend_object_iterator *iter TSRMLS_DC)
{
	frozen_array_it *iterator = (frozen_array_it*)(iter);
	frozen_array_object *obj = iterator->object;

	zend_hash_internal_pointer_reset_ex(Z_ARRVAL_P(obj->data), &(iterator->pos));
}

zend_object_iterator_funcs frozen_array_it_funcs = {
	frozen_array_it_dtor,
	frozen_array_it_valid,
	frozen_array_it_get_current_data,
	frozen_array_it_get_current_key,
	frozen_array_it_move_forward,
	frozen_array_it_rewind
};

zend_object_iterator* frozen_array_get_iterator(zend_class_entry *ce, zval *object, int by_ref TSRMLS_DC) 
{
	frozen_array_it  *iterator;
	frozen_array_object *obj = (frozen_array_object*)zend_object_store_get_object(object TSRMLS_CC);
	
	if(by_ref)
	{
		zend_error(E_ERROR, "An iterator cannot be used with foreach by reference");
	}
	
	iterator = emalloc(sizeof(frozen_array_it));

	Z_ADDREF_P(object);

	iterator->intern.it.data = (void*)object;
	iterator->intern.it.funcs = &frozen_array_it_funcs;
	iterator->intern.ce = ce;
	iterator->intern.value = NULL;
	iterator->object = obj;
	iterator->current = NULL;
	
	return  (zend_object_iterator*)iterator;
}

/* }}} */

void frozen_array_init(TSRMLS_D)
{
	zend_class_entry ce;

	/* change the ones we need to */

	INIT_CLASS_ENTRY(ce, "FrozenArray", frozen_array_functions);
	ce.create_object = frozen_array_new;
	ce.get_iterator = frozen_array_get_iterator;

	frozen_array_ce = zend_register_internal_class(&ce TSRMLS_CC);
	
	memcpy(&frozen_array_object_handlers, 
			zend_get_std_object_handlers(), 
			sizeof(zend_object_handlers));

	zend_class_implements(frozen_array_ce TSRMLS_CC, 1, zend_ce_arrayaccess);
}

void frozen_array_shutdown(TSRMLS_D)
{
	frozen_array_ce = NULL;
}

#ifndef IS_CONSTANT_TYPE_MASK
#define IS_CONSTANT_TYPE_MASK (~IS_CONSTANT_INDEX)
#endif

static HashTable* frozen_array_copy_hashtable(HashTable* dst, HashTable* src)
{
	Bucket* curr = NULL;
	Bucket* prev = NULL;
	Bucket* newp = NULL;
	int first = 1;

    if (!dst)
	{
        dst = (HashTable*) pemalloc(sizeof(HashTable), 1);
    }

	memcpy(dst, src, sizeof(HashTable));

	dst->arBuckets = pemalloc(sizeof(Bucket*) * dst->nTableSize, 1);
	dst->pDestructor = NULL;

    memset(dst->arBuckets, 0, dst->nTableSize * sizeof(Bucket*));
    dst->pInternalPointer = NULL;
    dst->pListHead = NULL;

    for (curr = src->pListHead; curr != NULL; curr = curr->pListNext)
	{
        int n = curr->h % dst->nTableSize;

		newp = pemalloc(sizeof(Bucket) + curr->nKeyLength - 1, 1);
		memcpy(newp, curr, sizeof(Bucket) + curr->nKeyLength - 1);
		
		
        /* insert 'newp' into the linked list at its hashed index */
        if (dst->arBuckets[n])
		{
            newp->pNext = dst->arBuckets[n];
            newp->pLast = NULL;
            newp->pNext->pLast = newp;
        }
        else
		{
            newp->pNext = newp->pLast = NULL;
        }

        dst->arBuckets[n] = newp;

        /* copy the bucket data */
		newp->pDataPtr = frozen_array_copy_zval_ptr(NULL, curr->pDataPtr);
		newp->pData = &newp->pDataPtr;

        /* insert 'newp' into the table-thread linked list */
        newp->pListLast = prev;
        newp->pListNext = NULL;

        if (prev) {
            prev->pListNext = newp;
        }

        if (first) {
            dst->pListHead = newp;
            first = 0;
        }

        prev = newp;
    }

    dst->pListTail = newp;

    return dst;
}

zval* frozen_array_copy_zval_ptr(zval* dst, zval* src)
{
	if(!dst)
	{
		dst = pemalloc(sizeof(zval), 1);
	}

	memcpy(dst, src, sizeof(zval));

    switch (src->type & IS_CONSTANT_TYPE_MASK)
	{
		case IS_RESOURCE:
		case IS_BOOL:
		case IS_LONG:
		case IS_DOUBLE:
		case IS_NULL:
			break;

		case IS_CONSTANT:
		case IS_STRING:
		{
			if (Z_STRVAL_P(src))
			{
				Z_STRVAL_P(dst) = pemalloc(sizeof(char)*Z_STRLEN_P(src)+1, 1);
				memcpy(Z_STRVAL_P(dst), Z_STRVAL_P(src), Z_STRLEN_P(src)+1);
			}
        }
        break;

		case IS_ARRAY:
		case IS_CONSTANT_ARRAY:
			Z_ARRVAL_P(dst) = frozen_array_copy_hashtable(NULL, Z_ARRVAL_P(src));
			break;

		case IS_OBJECT:
			/* ERROR ! */
			dst->type = IS_NULL;
			break;

		default:
			assert(0);
	}

    return dst;
}

static void frozen_array_free_hashtable(HashTable **_ht)
{
	HashTable *ht = *_ht;
	Bucket* curr = ht->pListHead;
	Bucket* tmp = NULL;
	
    while(curr)
	{
		tmp = curr;
		curr = curr->pListNext;
		frozen_array_free_zval_ptr((zval**)(&tmp->pDataPtr));
		pefree(tmp, 1);
	}

	pefree(ht->arBuckets, 1);
	pefree(ht, 1);
	*_ht = NULL;
}

void frozen_array_free_zval_ptr(zval** val)
{
	zval *value = *val;

    switch (value->type & IS_CONSTANT_TYPE_MASK)
	{
		case IS_RESOURCE:
		case IS_BOOL:
		case IS_LONG:
		case IS_DOUBLE:
		case IS_NULL:
			break;

		case IS_CONSTANT:
		case IS_STRING:
		{
			if (Z_STRVAL_P(value))
			{
				pefree(Z_STRVAL_P(value), 1);
			}
        }
        break;

		case IS_ARRAY:
		case IS_CONSTANT_ARRAY:
			frozen_array_free_hashtable(&Z_ARRVAL_P(value));
			break;

		case IS_OBJECT:
			/* ERROR ! */
			break;

		default:
			assert(0);
	}

	pefree(value, 1);
	*val = NULL;
}

zval* frozen_array_unserialize(const char* filename TSRMLS_DC)
{
	zval* data;
	zval* retval;
	long len = 0;
	struct stat sb;
	char *contents, *tmp;
	FILE *fp;
	php_unserialize_data_t var_hash;

	if(VCWD_STAT(filename, &sb) == -1) 
	{
		return NULL;
	}

	fp = fopen(filename, "rb");

	len = sizeof(char)*sb.st_size;

	tmp = contents = malloc(len);

	fread(contents, 1, len, fp);

	MAKE_STD_ZVAL(data);

	PHP_VAR_UNSERIALIZE_INIT(var_hash);
	
	/* I wish I could use json */
	if(!php_var_unserialize(&data, (const unsigned char**)&tmp, contents+len, &var_hash TSRMLS_CC))
	{
		zval_ptr_dtor(&data);
		free(contents);
		fclose(fp);
		return NULL;
	}

	PHP_VAR_UNSERIALIZE_DESTROY(var_hash);

	retval = frozen_array_copy_zval_ptr(NULL, data);

	zval_ptr_dtor(&data);
	free(contents);
	fclose(fp);

	return retval;
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim>600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
